////////////////////////////////////////////////////////////////////// 
// mach2dat/LogisticRegression.h 
// (c) 2008 Yun Li
// 
// March 15, 2008
//

#include "LogisticRegression.h"
#include "MathSVD.h"
#include "MathCholesky.h"
#include "StringHash.h"
#include "MathStats.h"

LogisticRegression::LogisticRegression()
{
}

LogisticRegression::~LogisticRegression()
{
}

double LogisticRegression::GetDeviance(Matrix & X, Vector & y)
{
	double ll = 0.0;
	for (int i = 0; i < X.rows; i++)
	{
		double t = 0.0;
		for (int j = 0; j < X.cols; j++)
			t += B[j] * X[i][j];
		double yhat = 1 / (1 + exp(-t));

		ll += y[i] == 1 ? log (yhat) : log(1 - yhat);
	}

	double deviance = -2.0 * ll;
	return deviance;
}


bool LogisticRegression::FitLogisticModel(Matrix & X, Vector & y, int nrrounds)
{
	B.Dimension(X.cols);
	B.Zero();

	covB.Dimension(X.cols, X.cols);
	covB.Zero();

	Vector p, V;
	p.Dimension(X.rows);
	V.Dimension(X.rows);
	p.Zero();
	V.Zero();

	int rounds = 0;

	// Newton-Raphson
	while (rounds < nrrounds)
	//while (true)
	{
		// beta = beta + solve(t(X)%*%diag(p*(1-p))%*%X)%*%t(X)%*%(Y-p);
		for (int i = 0; i < X.rows; i++)
		{
			double d = 0;
			for (int k = 0; k < X.cols; k ++)
			{
				d += B[k] * X[i][k];
			} // for parameter beta-k

			p[i] = 1.0 / (1.0 + exp(-d));
			V[i] = p[i] * (1 - p[i]);
		} // for observation i

		// The first part: solve / inverse
		Matrix D;
		D.Dimension(X.cols, X.cols);
		D.Zero();

		for (int k = 0; k < X.cols; k++)
		{
			for (int l = k; l < X.cols; l++)
			{
				double Dentry = 0.0;
				for (int i = 0; i < X.rows; i++)
					Dentry += X[i][k] * V[i] * X[i][l];
				D[k][l] = D[l][k] = Dentry;
			}
		}

		Matrix Dinv;
		Dinv.Dimension(X.cols, X.cols);
		Dinv.Zero();
		Dinv = D;
		// SVD svd;
		// svd.SVDInverse(Dinv);
		Cholesky chol;
		if (!chol.TryDecompose(D))
			return false;
		chol.Decompose(D);
		chol.Invert();
		Dinv = chol.inv;

		// Accumulate up to the second part: multiply by t(X)
		Matrix Dtwo;
		Dtwo.Dimension(X.cols, X.rows);
		Dtwo.Zero();

		for (int k = 0; k < X.cols; k++)
			for (int i = 0; i < X.rows; i++)
				for (int l = 0; l < X.cols; l++)
					Dtwo[k][i] += Dinv[k][l] * X[i][l];

		// The last part, residuals
		Vector residuals;
		residuals.Dimension(X.rows);
		residuals.Zero();
		for (int i = 0; i < X.rows; i++)
			residuals[i] = y[i] - p[i];

		Vector deltaB;
		deltaB.Dimension(X.cols);
		deltaB.Zero();
		for (int k = 0; k < X.cols; k++)
			for (int i = 0; i < X.rows; i++)
				deltaB[k] += Dtwo[k][i] * residuals[i];

		// update beta's and check for convergence
		double delta = 0.0;
		for (int k = 0; k < X.cols; k++)
		{
			delta += fabs(deltaB[k]);
			B[k] += deltaB[k];
		}

		//printf ("%d %-11.4g\n", rounds, delta);

		if (delta < 1e-6)
		{
			rounds = 0;
			break;
		}

		rounds ++;
	} // Newton-Raphson iterations

	if (rounds == nrrounds)
		return false;


	// obtain covariance matrix to perform Wald test
	// covB = solve(t(X)%*%V%*%X)
	
	// Transpose X and multiply by diagonal V
	Matrix XtV;
	XtV.Dimension(X.cols, X.rows);
	XtV.Zero();
	for (int i = 0; i < X.rows; i++)
		for (int k = 0; k < X.cols; k++)
			XtV[k][i] = X[i][k] * V[i];

	Matrix XtVX;
	XtVX.Dimension(X.cols, X.cols);
	XtVX.Zero();
	XtVX.Product(XtV,X);

	covB.Zero();
	covB = XtVX;
	//SVD svd;
	//svd.SVDInverse(covB);
	Cholesky chol;
	if (!chol.TryDecompose(covB))
		return false;
	chol.Decompose(covB);
	chol.Invert();
	covB = chol.inv;
	
	return true;
}



 
